from typing import Any, Generic, List, Optional, Tuple, TypeVar

from chex import Numeric
from optax import ScaleByAdamState, ScaleState, scale, scale_by_adam

from .annotations import PyTree
from .dataclass import dataclass

Parameters = TypeVar('Parameters', bound=PyTree)
State = TypeVar('State', bound=PyTree)


__all__ = ['GradientTransformation', 'ChainedGradientTransformation', 'Scale', 'ScaleByAdam',
           'adam']


@dataclass
class GradientTransformation(Generic[State, Parameters]):

    def init(self, parameters: Parameters) -> State:
        raise NotImplementedError

    def update(self,
               gradient: Parameters,
               state: State,
               parameters: Optional[Parameters]) -> Tuple[Parameters, State]:
        raise NotImplementedError


@dataclass
class ChainedGradientTransformation(GradientTransformation[List[PyTree], Parameters],
                                    Generic[Parameters]):

    transforms: List[GradientTransformation[Any, Parameters]]

    def init(self, parameters: Parameters) -> List[PyTree]:
        return [transform.init(parameters)
                for transform in self.transforms]

    def update(self,
               gradient: Parameters,
               state: List[PyTree],
               parameters: Optional[Parameters]) -> Tuple[Parameters, List[PyTree]]:
        new_state: List[PyTree] = []
        for sub_state, transform in zip(state, self.transforms):
            gradient, new_state = transform.update(gradient, sub_state, parameters)
            new_state.append(new_state)
        return gradient, new_state


@dataclass
class Scale(GradientTransformation[ScaleState, Parameters], Generic[Parameters]):

    step_size: Numeric

    def init(self, parameters: Parameters) -> ScaleState:
        return scale(self.step_size).init(parameters)

    def update(self,
               gradient: Parameters,
               state: ScaleState,
               parameters: Optional[Parameters]) -> Tuple[Parameters, ScaleState]:
        return scale(self.step_size).update(gradient, state, parameters)


@dataclass
class ScaleByAdam(GradientTransformation[ScaleByAdamState, Parameters], Generic[Parameters]):

    beta1: Numeric = 0.9
    beta2: Numeric = 0.999
    epsilon: Numeric = 1e-8
    epsilon_root: Numeric = 0.0

    def init(self, parameters: Parameters) -> ScaleByAdamState:
        return scale_by_adam(self.beta1, self.beta2, self.epsilon, self.epsilon_root).init(
            parameters)

    def update(self,
               gradient: Parameters,
               state: ScaleByAdamState,
               parameters: Optional[Parameters]) -> Tuple[Parameters, ScaleByAdamState]:
        return scale_by_adam(self.beta1, self.beta2, self.epsilon, self.epsilon_root).update(
            gradient, state, parameters)


def adam(learning_rate: float,
         beta1: float = 0.9,
         beta2: float = 0.999,
         epsilon: float = 1e-8) -> GradientTransformation[Any, Any]:
    return ChainedGradientTransformation([ScaleByAdam(beta1, beta2, epsilon),
                                          Scale(-learning_rate)])



import sys
from numbers import Number
from typing import Any, Type, Union

import colorful as cf
import networkx as nx
import numpy as np
from jax.interpreters.ad import JVPTracer
from jax.interpreters.batching import BatchTracer
from jax.interpreters.partial_eval import JaxprTracer
from jax.interpreters.xla import DeviceArray

__all__ = ['Displayable', 'print_generic']


if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable

    @runtime_checkable
    class Displayable(Protocol):
        """
        This protocol identifies classes that support the `display_generic` mechanism.
        """
        def display(self, show_values: bool = True, indent: int = 0) -> str:
            ...
else:
    class Displayable:
        pass


def print_generic(*args: Any, **kwargs: Any) -> None:
    for value in args:
        print(display_generic(value))
    for key, value in kwargs.items():
        print(display_key_and_value(key, value, "=", True, 0))


def display_generic(value: Any, show_values: bool = True, indent: int = 0) -> str:
    # pylint: disable=too-many-return-statements
    if isinstance(value, JVPTracer):
        return cf.magenta(f"JVPTracer {value.shape} {value.dtype}") + "\n"
    if isinstance(value, JaxprTracer):
        return cf.magenta(f"JaxprTracer {value.shape} {value.dtype}") + "\n"
    if isinstance(value, BatchTracer):
        return cf.magenta(f"BatchTracer {value.shape} {value.dtype} "
                          f"batched over {value.val.shape[value.batch_dim]}") + "\n"
    if isinstance(value, (np.ndarray, DeviceArray)):
        if isinstance(value, np.ndarray):
            retval = cf.yellow(f"NumPy Array {value.shape}") + "\n"
        elif isinstance(value, DeviceArray):
            retval = cf.violet(f"Jax Array {value.shape}") + "\n"
        if show_values:
            retval += _show_array(indent + 1, value)
        return retval
    if isinstance(value, Number) or value is None:
        return cf.cyan(str(value)) + "\n"
    if (isinstance(value, Displayable)
            if sys.version_info >= (3, 8)
            else hasattr(value, 'display')):
        return value.display(show_values=show_values, indent=indent)
    if isinstance(value, dict):
        return (display_class(dict)
                + "".join(display_key_and_value(key, sub_value, "=", show_values, indent)
                          for key, sub_value in value.items()))
    if isinstance(value, (tuple, list)):
        return (display_class(type(value))
                + "".join(display_key_and_value("", sub_value, "", show_values, indent)
                          for sub_value in value))
    if isinstance(value, nx.Graph):
        return display_graph(value, show_values, indent)
    return cf.red(str(value)) + "\n"


def display_graph(graph: nx.Graph, show_values: bool, indent: int = 0) -> str:
    directed = isinstance(graph, nx.DiGraph)
    arrow = cf.base00('⟶  ' if directed else '🡘 ')
    retval = display_class(type(graph))
    for name, value in graph.nodes.items():
        retval += display_key_and_value(name, value['element'], ": ", show_values, indent)
    for (source, target), value in graph.edges.items():
        key = f"{source}{arrow}{target}"
        retval += display_key_and_value(key, value['link'], ": ", show_values, indent)
    return retval


def display_class(cls: Type[Any]) -> str:
    return cf.orange(cls.__name__) + "\n"


def display_key_and_value(
        key: str, value: Any, separator: str, show_values: bool = True, indent: int = 0) -> str:
    return (_indent_space(indent + 1) + cf.blue(key) + cf.base00(separator)
            + display_generic(value, show_values, indent + 1))


# Private functions --------------------------------------------------------------------------------
_indentation = 4


def _indent_space(indent: int) -> str:
    return (indent * _indentation) * " "


def _format_number(x: float) -> str:
    return f"{x:10.4f}"


def _show_array(indent: int, array: Union[np.ndarray, DeviceArray]) -> str:
    if len(array.shape) == 0:
        return _indent_space(indent) + _format_number(array[()]) + "\n"
    if np.prod(array.shape) == 0:
        return ""
    if 1 in array.shape:
        return _show_array(indent, array[tuple(0 if s == 1 else slice(None)
                                               for s in array.shape)])
    if len(array.shape) == 1:
        return _indent_space(indent) + "  ".join(_format_number(array[i])
                                                 for i in range(array.shape[0])) + "\n"
    if len(array.shape) == 2:
        return "".join(_show_array(indent, array_slice)
                       for array_slice in array)
    return ""
